// weather.cc -- Davis GroWeather logging software
// Copyright (C) 1998 Ari Mujunen. (Ari.Mujunen@hut.fi)

// This is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License.
// See the file 'COPYING' for details.

// $Log: weather.cc,v $
// Revision 1.7  1998/12/01 11:21:50  amn
// Now compiles without GNU g++ "old" libraries both on 'bo' and 'hamm'.
//
// Revision 1.6  1998/09/04 10:41:38  amn
// Made the sleep between MDMPs longer, 100 sec.
//
// Revision 1.5  1998/05/14 13:43:23  amn
// Changed LWReg from 0x01to 0x04 (leaf wetness on).
//
// Revision 1.4  1998/05/11 09:52:40  amn
// 'Status' -> 'Stat'.
//
// Revision 1.3  1998/05/11 09:48:46  amn
// Added rain, wind, solar, power voltage.
//
// Revision 1.2  1998/05/10 19:13:48  amn
// Added flushing stdout after each readout.
//
// Revision 1.1  1998/05/10 18:58:19  amn
// Initial revision
//

#include "registration.h"
RCSID(weather_cc, "$Id: weather.cc,v 1.7 1998/12/01 11:21:50 amn Exp $");

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <limits.h>
#include <sys/time.h>  // for 'gettimeofday()' (needs unistd.h, too)
#include <time.h>  // for 'gmtime()', 'asctime()' et al

#include <assert.h>

#include "cSerialLine.hh"


#define MYNAME "weather: "


//-------

#include "ascii.h"

#include "ccittcrc.h"

#define DEBUG_IO 0
#define DEBUG_CRC 0
#define DEBUG_NIBBLES 0

class cWeatherStation {
private:
  cSerialLine sl;
  unsigned char mdmp_reply[258];
public:
  void init(const char *serial_device, const int baudRate = 2400);
  void cleanup(void);
  int getMdmp(void);
  int get(int nibbleOffset, int nibbles, const int isSigned);
  int sendCommand(const char *cmd);
};  // cWeatherStation

void
cWeatherStation::init(const char *serial_device, const int baudRate)
{
  this->sl.open(serial_device);
  // Can later use 'sl.whatPathname()' to access the stored serial device path.
  this->sl.serialSettings(/*baudRate=>*/ baudRate,
			  /*numberOfDataBits=>*/ 8,
			  /*numberOfStopBits=>*/ 1,
			  /*whichParity=>*/ eParityNone,
			  /*minBytes=>*/ 0,
			  /*maxTimeout=>*/ 100);  // 10 seconds
  // Davis GroWeather promises to respond to 'MDMP' in 6 seconds.
}  // init

void
cWeatherStation::cleanup(void)
{
  this->sl.close();
}  // cleanup

int
cWeatherStation::getMdmp(void)
{
  const int rcvCount = sizeof(this->mdmp_reply);
  int rcvBytes;
  
  // Send the message to the serial line.
  this->sl.flushPendingInput();
  this->sl.write("MDMP\r", 5);

  // Retrieve the ACK/'!' reply.
  rcvBytes = this->sl.read(this->mdmp_reply, 1);

  if (rcvBytes <= 0) {  // no response at all
    return 0;
  }

  if (this->mdmp_reply[0] != eAsciiAck) {
    // 'MDMP' not understood by weather station
#if DEBUG_IO
    fprintf(stderr, MYNAME "GroWeather at %s replied with '%02X' to 'MDMP'\n", this->sl.whatPathname(), this->mdmp_reply[0]);
#endif
    return 0;
  }

  // Retrieve the real 258 byte response.
  rcvBytes = this->sl.read(this->mdmp_reply, rcvCount);
#if DEBUG_IO
  fprintf(stderr, MYNAME "rcvBytes = %d\n", rcvBytes);
#endif

  if (rcvBytes <= 0) {  // no response at all
    return 0;
  }
  if (rcvBytes < rcvCount) {  // timeout because response was too short
    return rcvBytes;
  }

  // Must not have read over the end of buffer.
  assert ( rcvBytes == rcvCount );

  // Got the whole buffer, check CRC and return the number of bytes.
  {
    int crc = 0;
    unsigned i;

    for (i=0; i < sizeof(this->mdmp_reply); i++) {
      crc = 0xffff & (crc_table[(crc >> 8) ^ this->mdmp_reply[i] ] ^ (crc << 8));
    }
#if DEBUG_CRC
    fprintf(stderr, MYNAME "crc = %d\n", crc);
#endif
    if (crc != 0) {
      return 0;
    }
  }  // crc

  return rcvBytes;
}  // getMdmp

int
cWeatherStation::get(int nibbleOffset, int nibbles, const int isSigned)
{
  int bo = nibbleOffset / 2;
  int val = 0;
  int ns = 0;

  if ((nibbles <= 0) || (nibbles > 8)) {
    return INT_MIN;
  }
  if ((nibbleOffset < 0x0000) || (nibbleOffset > 0x01ff)) {
    return INT_MIN;
  }
  if ((nibbleOffset & 0x0001) != 0) {
    // Odd nibble address,  store least significant nibble.
    val = (this->mdmp_reply[bo] & 0xf0) >> 4;
#if DEBUG_NIBBLES
    printf("[odd n %d = %02X] ", bo, val);
#endif
    bo++;
    nibbles--;
    ns += 4;
  }
  while (nibbles >= 2) {
    val |= (this->mdmp_reply[bo] << ns);
#if DEBUG_NIBBLES
    printf("[n %d = %08X] ", bo, val);
#endif
    bo++;
    nibbles -= 2;
    ns += 8;
  }  // while two or more nibbles left
  if (nibbles >= 1) {
    val |= ( (this->mdmp_reply[bo] & 0x0f) << ns);
#if DEBUG_NIBBLES
    printf("[last n %d = %08X] ", bo, val);
#endif
    nibbles--;
    ns += 4;
  }
  assert( nibbles == 0 );

  // Take care of signed numbers.
  if (isSigned) {
    int maxnibbleint = (0x01) << (ns-1);

#if DEBUG_NIBBLES
    printf("[signed, is > %08X?]", maxnibbleint);
#endif
    if ((val == maxnibbleint) || (val == (maxnibbleint-1))) {
      // Negative or positive min/max indicates error/missing value
      // in weather station.
      return INT_MIN;
    }
    if (val > maxnibbleint) {
      maxnibbleint <<= 1;
#if DEBUG_NIBBLES
    printf("[yes, subtract %08X]", maxnibbleint);
#endif
      val -= maxnibbleint;
    }
  }
  return val;
}  // get


int
cWeatherStation::sendCommand(const char *cmd)
{
  int rcvBytes;
  char bu[1];

  // Send the message to the serial line.
  this->sl.flushPendingInput();
  this->sl.write(cmd, strlen(cmd));

  // Retrieve the ACK/'!' reply.
  rcvBytes = this->sl.read(bu, 1);

  if (rcvBytes <= 0) {  // no response at all
    return 1;
  }

  if (bu[0] != eAsciiAck) {
    // 'cmd' not understood by weather station
#if 1  //DEBUG_IO
    fprintf(stderr, MYNAME "GroWeather at %s replied with '%02X' to '%s'\n", this->sl.whatPathname(), bu[0], cmd);
#endif
    return 1;
  }
  return 0;
}  // sendCommand


//---
char *getUTC(char *p)
{
  struct timeval temptv;
  time_t temptime;
  struct tm *ti;
  static char bu[17];
  int centisec;

  if (p == NULL) {
    p = bu;
  }
  
  assert( gettimeofday(&temptv, NULL) == 0 );
  temptime = (time_t)temptv.tv_sec;
  ti = gmtime(&(temptime));
  centisec = temptv.tv_usec / 10000;
  if (centisec > 99) {
    centisec = 99;
  }
  assert( sprintf(p, "%04d%02d%02d%02d%02d%02d%02d",
          ti->tm_year + 1900,
          ti->tm_mon + 1,
          ti->tm_mday,
          ti->tm_hour,
          ti->tm_min,
          ti->tm_sec,
          centisec
                  // int     tm_wday;        /* day of the week */
                  // int     tm_yday;        /* day in the year */
                  // int     tm_isdst;       /* daylight saving time */
          ) <= ((int)sizeof(bu)-1) );
  return p;
}  //getUTC


#include <math.h>
#include <string.h>  // for 'strdup()'


int
main (int argc, char **argv)
{
  class cWeatherStation ws1;

  int option_char;

  // Configuration file / command line option defaults.
  // Command line wins, default conf file is next,
  // and these hard-coded values are the last resort.
  int my_baud_rate = 2400;
  char *my_serial_device = strdup("/dev/ttyS2");
  assert (my_serial_device != NULL);

  fprintf (stderr, MYNAME "$Revision: 1.7 $ $Date: 1998/12/01 11:21:50 $\n");

  // Process default conf file.
  // xxx process_conf(/*file=>*/ "/etc/weather.conf" );

  // Process every option character.
  while ((option_char = getopt(argc, argv, "c:d:b:")) != EOF)
    switch (option_char) {
       case 'c': {
         assert( optarg != NULL );
	 // xxx process_conf(/*file=>*/optarg);
	 break;
       }
       case 'd': {
	 free(my_serial_device);
	 assert ((my_serial_device = strdup(optarg)) != NULL);
         break;
       }
       case 'b': {
	 my_baud_rate = atoi (optarg);
	 break;
       }
       case ':':  // missing parameter after a option letter
       case '?': {
	 fprintf (stderr,
		  "usage: %s [-d<serial device>] [-b<baud>]\n", argv[0]);
	 exit(EXIT_FAILURE);
       }
    }  // switch

  ws1.init(my_serial_device, my_baud_rate);

  while (1) {
    double tp2C, relH, barMbar, /**dewC,**/ dewCalcC;
    double rainRate, leafWetness, solarRad, windSpeed, windDir, voltage;
    char lastUTC[17];  // yyyymmddhhmmssss+\0

    if (ws1.getMdmp() != 258) {
      printf("%s\tStat\t1\n", getUTC(NULL));
      fflush(stdout);
      sleep(2);
      continue;
    }
    getUTC(lastUTC);
    printf("%s\tStat\t0\n", lastUTC);

    // Model number, GroWeather == 4.
    {
      int modelCode;

      if ((modelCode = ws1.get(0x004d,1,0)) != 4) {
	fprintf(stderr, MYNAME "station model code %d not == 4, GroWeather\n",
		modelCode);
      }
    }

    // Select leaf wetness sensor in soil temp interface (LWReg <- 0x04).
    if (ws1.get(0x000e,1,0) != 0x04) {
      static const char leafcmd[8] = { 'W', 'W', 'R', 0x11, 0x0e, 0x04, '\r' };

      fprintf(stderr, MYNAME "forcing leaf wetness sensor on...\n");
      // Write to station memory bank 0 offset 0x0e nibble value 1.
      if (ws1.sendCommand(leafcmd)) {
	fprintf(stderr, MYNAME "forcing leaf wetness failed.\n");
      }
    }
  
    // Tp2Dta in Fahrenheit --> C.
    // (Tp2Cal calibration "numbers" (offsets) are usually zero in GroWeather.)
    tp2C = ((double)ws1.get(0x0142,3,1)/10.0 - 32.0) * (5.0/9.0);
    printf("%s\tOutT\t%g\n", lastUTC, tp2C);

    // Hm2Dta 1--100%.
    // (Hm2Cal calibration "numbers" (offsets) are usually zero in GroWeather.)
    relH = (double)ws1.get(0x0173,2,1);
    if (relH < 1.0) {
      relH = 1.0;
    }
    if (relH > 100.0) {
      relH = 100.0;
    }
    printf("%s\tRelH\t%g\n", lastUTC, relH);

  // BarDta in 0.001 inHg --> mbar.
  // (inHg -> mbar by GNU 'units'.)
  // (BarFlag indicates steady/rise/fall, can be deduced otherwise.)
    barMbar = (double)ws1.get(0x0100,4,0) * 0.033863886;
    printf("%s\tBaro\t%g\n", lastUTC, barMbar);

#if 0
    // We omit dewpoint 'DewDta' calculated by station.
    dewC = ((double)ws1.get(0x008a,3,1)/10.0 - 32.0) * (5.0/9.0);
    printf("%s\tDewP\t%g\n", lastUTC, dewC);
#endif

    // Calculated dewpoint in C.
    {
      double ews;

      if (/* relH 1..100 already */ tp2C < -50.0 /*???*/ || tp2C > 70.0) {
	printf("%s\tDewP\tNaN\n", lastUTC);
      } else {
	ews = log( relH * 0.01 * exp( (17.502*tp2C) / (240.9+tp2C) ) );
#if 0
	printf("ews = %g, ", ews);
#endif
	dewCalcC = (240.9*ews) / (17.5-ews);
	printf("%s\tDewP\t%g\n", lastUTC, dewCalcC);
      }
    }

    // RateDta, rain rate (half seconds since last "click").
    rainRate = (double)ws1.get(0x01d3,3,0);
    printf("%s\tRain\t%g\n", lastUTC, rainRate);

  // LeafDta, leaf wetness 0--15.
    leafWetness = (double)ws1.get(0x01de,1,0);
    printf("%s\tLeaf\t%g\n", lastUTC, leafWetness);

  // SradDta, solar radiation (watts / square meter)
    solarRad = (double)ws1.get(0x01b8,3,0);
    printf("%s\tSola\t%g\n", lastUTC, solarRad);

  // SpdDta, wind speed MPH --> m/s (by GNU 'units').
    windSpeed = (double)ws1.get(0x0060,2,0) * 0.44704;
    printf("%s\tWind\t%g\n", lastUTC, windSpeed);

  // DirFlags bit 0x04 --> direction bad, otherwise DirDta degrees.
    if (ws1.get(0x003d,1,0) & 0x04) {
      windDir = (double)(INT_MIN);
    } else {
      windDir = (double)ws1.get(0x018b,2,0);
    }
    printf("%s\tWDir\t%g\n", lastUTC, windDir);

  // PowVolt, power supply voltage.
  // (PwrFlags bit 0x08 (prim. power low) --> less than 9???V.)
    if (ws1.get(0x007b,1,0) & 0x08) {
      voltage = 6.0;  // xxx
    } else {
      voltage = 10.25 + ws1.get(0x01dd,1,0)*0.25;
    }
    printf("%s\tVolt\t%g\n", lastUTC, voltage);

#if 0
    //   offset # 0==unsigned/1=signed
    //         nibbles
    DOIT(0x0075,1,0,"%01X",HiDir);  // direction of station-detected hi wind
    DOIT(0x0194,6,0,"%06X",Time);  // (wrong :-) station time
    DOIT(0x019e,5,0,"%05X",Date);  // (wrong :-) station date
    DOIT(0x0036,5,0,"%05X",AlarmReg);  // station hi-lo alarms...
    DOIT(0x003b,1,0,"%01X",AOMStat);  // comm err seems to be always 4==bad...
#endif

    // Before the next 'MDMP' let's wait a little to allow GroWeather
    // to do something on its own.
    fflush(stdout);
    sleep(100);
  }  // while forever

  ws1.cleanup();
}  // main
